import argh
from gin_train.cli.gin_train import gin_train

# logging
import pkg_resources
import logging
import logging.config
logging.config.fileConfig(pkg_resources.resource_filename(__name__, "logging.conf"))
logger = logging.getLogger(__name__)


def main():
    # assembling:
    # parser = argh.ArghParser()
    # parser.add_commands([gin_train])
    # argh.dispatch(parser)
    argh.dispatch_command(gin_train)



"""Train models using gin configuration
"""
import gin
import json
import sys
import os
import yaml
from uuid import uuid4
from fs.osfs import OSFS
import numpy as np
from tqdm import tqdm
from gin_train.remote import upload_dir
from gin_train.config import create_tf_session
from gin_train.utils import write_json, Logger, NumpyAwareJSONEncoder, prefix_dict
from comet_ml import Experiment

try:
    import wandb
except ImportError:
    wandb = None
# import all modules registering any gin configurables

# configurables import
from gin_train.trainers import KerasTrainer
from gin_train import metrics
from gin_train import trainers
from gin_train import samplers


import logging
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def log_gin_config(output_dir, cometml_experiment=None, wandb_run=None):
    gin_config_str = gin.operative_config_str()

    print("Used config: " + "-" * 40)
    print(gin_config_str)
    print("-" * 52)
    with open(os.path.join(output_dir, "config.gin"), "w") as f:
        f.write(gin_config_str)
    # parse the gin config string to dictionary
    gin_config_str = "\n".join([x for x in gin_config_str.split("\n")
                                if not x.startswith("import")])
    gin_config_dict = yaml.load(gin_config_str
                                .replace("@", "")
                                .replace(" = %", ": ")
                                .replace(" = ", ": "))
    write_json(gin_config_dict,
               os.path.join(output_dir, "config.gin.json"),
               sort_keys=True,
               indent=2)

    if cometml_experiment is not None:
        # Skip any rows starting with import
        cometml_experiment.log_multiple_params(gin_config_dict)

    if wandb_run is not None:
        # This allows to display the metric on the dashboard
        wandb_run.config.update({k.replace(".", "/"): v for k, v in gin_config_dict.items()})


def add_file_logging(output_dir, logger, name='stdout'):
    os.makedirs(os.path.join(output_dir, 'log'), exist_ok=True)
    log = Logger(os.path.join(output_dir, 'log', name + '.log'), 'a+')  # log to the file
    fh = logging.FileHandler(os.path.join(output_dir, 'log', name + '.log'), 'a+')
    fh.setFormatter(logging.Formatter('[%(asctime)s] - [%(levelname)s] - %(message)s'))
    fh.setLevel(logging.INFO)
    logger.addHandler(fh)
    return log


def kv_string2dict(s):
    """Convert a key-value string: k=v,k2=v2,... into a dictionary
    """
    return yaml.load(s.replace(",", "\n").replace("=", ": "))


@gin.configurable
def train(output_dir,
          model=gin.REQUIRED,
          data=gin.REQUIRED,
          eval_metric=None,
          eval_train=False,
          eval_skip=[],
          trainer_cls=KerasTrainer,
          # shared
          batch_size=256,
          num_workers=8,
          # train-specific
          epochs=100,
          early_stop_patience=4,
          train_epoch_frac=1.0,
          valid_epoch_frac=1.0,
          train_samples_per_epoch=None,
          validation_samples=None,
          train_batch_sampler=None,
          stratified_sampler_p=None,
          tensorboard=True,
          seed=None,
          remote_dir='',
          cometml_experiment=None,
          wandb_run=None
          ):
    """Main entry point to configure in the gin config
    Args:
      model: compiled keras model
      data: tuple of (train, valid) Datasets
      eval_train: if True, also compute the evaluation metrics for the final model
        on the training set
      eval_skip List[str]: datasets to skip during evaluation
      seed: random seed to use (in numpy and tensorflow)
    """
    # from this point on, no configurable should be added. Save the gin config
    log_gin_config(output_dir, cometml_experiment, wandb_run)

    train_dataset, valid_dataset = data[0], data[1]

    if seed is not None:
        # Set the random seed
        import random
        random.seed(seed)
        np.random.seed(seed)
        try:
            import tensorflow as tf
            tf.set_random_seed(seed)
        except Exception:
            logger.info("Unable to set random seed for tensorflow")

    # make sure the validation dataset names are unique
    if isinstance(valid_dataset, list):
        dataset_names = []
        for d in valid_dataset:
            dataset_name = d[0]
            if dataset_name in dataset_names:
                raise ValueError("The dataset names are not unique")
            dataset_names.append(dataset_name)

    if stratified_sampler_p is not None and train_batch_sampler is not None:
        raise ValueError("stratified_sampler_p and train_batch_sampler are mutually exclusive."
                         " Please specify only one of them.")

    if stratified_sampler_p is not None and train_batch_sampler is None:
        # HACK - there is no guarantee that train_dataset.get_targets() will exist
        # Maybe we have to introduce a ClassificationDataset instead which will
        # always implement get_targets()
        logger.info(f"Using stratified samplers with p: {stratified_sampler_p}")
        train_batch_sampler = samplers.StratifiedRandomBatchSampler(train_dataset.get_targets().max(axis=1),
                                                                    batch_size=batch_size,
                                                                    p_vec=stratified_sampler_p,
                                                                    verbose=True)

    tr = trainer_cls(model, train_dataset, valid_dataset, output_dir, cometml_experiment, wandb_run)
    tr.train(batch_size=batch_size,
             epochs=epochs,
             early_stop_patience=early_stop_patience,
             num_workers=num_workers,
             train_epoch_frac=train_epoch_frac,
             valid_epoch_frac=valid_epoch_frac,
             train_samples_per_epoch=train_samples_per_epoch,
             validation_samples=validation_samples,
             train_batch_sampler=train_batch_sampler,
             tensorboard=tensorboard)
    final_metrics = tr.evaluate(eval_metric, batch_size=batch_size, num_workers=num_workers,
                                eval_train=eval_train, eval_skip=eval_skip, save=True)
    # pass
    logger.info("Done!")
    print("-" * 40)
    print("Final metrics: ")
    print(json.dumps(final_metrics, cls=NumpyAwareJSONEncoder, indent=2))

    # upload files to a custom remote directory
    if remote_dir:
        logger.info("Uploading files to: {}".format(remote_dir))
        upload_dir(output_dir, remote_dir)

    # upload files to comet.ml
    if cometml_experiment is not None:
        logger.info("Uploading files to comet.ml")
        for f in tqdm(list(OSFS(output_dir).walk.files())):
            # [1:] removes trailing slash
            cometml_experiment.log_asset(file_path=os.path.join(output_dir, f[1:]),
                                         file_name=f[1:])
    return final_metrics


def gin_train(gin_files,
              output_dir,
              gin_bindings='',
              gpu=0,
              memfrac=0.45,
              framework='tf',
              cometml_project="",
              wandb_project="",
              remote_dir="",
              run_id=None,
              note_params="",
              force_overwrite=False):
    """Train a model using gin-config

    Args:
      gin_file: comma separated list of gin files
      output_dir: where to store the results. Note: a subdirectory `run_id`
        will be created in `output_dir`.
      gin_bindings: comma separated list of additional gin-bindings to use
      gpu: which gpu to use. Example: gpu=1
      memfrac: what fraction of the GPU's memory to use
      framework: which framework to use. Available: tf
      cometml_project: comet_ml project name. Example: Avsecz/basepair.
        If not specified, cometml will not get used
      wandb_project: wandb `<entity>/<project>` name. Example: Avsecz/test.
        If not specified, wandb will not be used
      remote_dir: additional path to the remote directory. Can be an s3 path.
        Example: `s3://mybucket/model1/exp1`
      run_id: manual run id. If not specified, it will be either randomly
        generated or re-used from wandb or comet.ml.
      note_params: take note of additional key=value pairs.
        Example: --note-params note='my custom note',feature_set=this
      force_overwrite: if True, the output directory will be overwritten
    """

    sys.path.append(os.getcwd())
    if cometml_project:
        logger.info("Using comet.ml")
        workspace, project_name = cometml_project.split("/")
        cometml_experiment = Experiment(project_name=project_name, workspace=workspace)
        # TODO - get the experiment id
        # specify output_dir to that directory
    else:
        cometml_experiment = None

    if wandb_project:
        assert "/" in wandb_project
        entity, project = wandb_project.split("/")
        if wandb is None:
            logger.warn("wandb not installed. Not using it")
            wandb_run = None
        else:
            wandb._set_stage_dir("./")  # Don't prepend wandb to output file
            if run_id is not None:
                wandb.init(project=project,
                           dir=output_dir,
                           entity=entity,
                           resume=run_id)
            else:
                # automatically set the output
                wandb.init(project=project,
                           entity=entity,
                           dir=output_dir)
            wandb_run = wandb.run
            logger.info("Using wandb")
            print(wandb_run)
    else:
        wandb_run = None

    # update the output directory
    if run_id is None:
        if wandb_run is not None:
            run_id = os.path.basename(wandb_run.dir)
        elif cometml_experiment is not None:
            run_id = cometml_experiment.id
        else:
            # random run_id
            run_id = str(uuid4())
    output_dir = os.path.join(output_dir, run_id)
    if remote_dir:
        remote_dir = os.path.join(remote_dir, run_id)
    if wandb_run is not None:
        # make sure the output directory is the same
        # wandb_run._dir = os.path.normpath(output_dir)  # This doesn't work
        # assert os.path.normpath(wandb_run.dir) == os.path.normpath(output_dir)
        # TODO - fix this assertion-> the output directories should be the same
        # in order for snakemake to work correctly
        pass
    # -----------------------------

    if os.path.exists(os.path.join(output_dir, 'config.gin')):
        if force_overwrite:
            logger.info(f"config.gin already exists in the output "
                        "directory {output_dir}. Removing the whole directory.")
            import shutil
            shutil.rmtree(output_dir)
        else:
            raise ValueError(f"Output directory {output_dir} shouldn't exist!")
    os.makedirs(output_dir, exist_ok=True)  # make the output directory. It shouldn't exist

    # add logging to the file
    add_file_logging(output_dir, logger)

    if framework == 'tf':
        import gin.tf
        if gpu is not None:
            logger.info(f"Using gpu: {gpu}, memory fraction: {memfrac}")
            create_tf_session(gpu, per_process_gpu_memory_fraction=memfrac)

    gin.parse_config_files_and_bindings(gin_files.split(","),
                                        bindings=gin_bindings.split(";"),
                                        skip_unknown=False)

    # write note_params.json
    if note_params:
        logger.info(f"note_params: {note_params}")
        note_params_dict = kv_string2dict(note_params)
    else:
        note_params_dict = dict()
    write_json(note_params_dict,
               os.path.join(output_dir, "note_params.json"),
               sort_keys=True,
               indent=2)

    # comet - log environment
    if cometml_experiment is not None:
        # log other parameters
        cometml_experiment.log_multiple_params(dict(gin_files=gin_files,
                                                    gin_bindings=gin_bindings,
                                                    output_dir=output_dir,
                                                    gpu=gpu), prefix='cli/')
        cometml_experiment.log_multiple_params(note_params_dict)

        exp_url = f"https://www.comet.ml/{cometml_experiment.workspace}/{cometml_experiment.project_name}/{cometml_experiment.id}"
        logger.info("Comet.ml url: " + exp_url)
        # write the information about comet.ml experiment
        write_json({"url": exp_url,
                    "key": cometml_experiment.id,
                    "project": cometml_experiment.project_name,
                    "workspace": cometml_experiment.workspace},
                   os.path.join(output_dir, "cometml.json"),
                   sort_keys=True,
                   indent=2)

    # wandb - log environment
    if wandb_run is not None:
        write_json({"url": wandb_run.get_url(),
                    "key": wandb_run.id,
                    "project": wandb_run.project,
                    "path": wandb_run.path,
                    "group": wandb_run.group
                    },
                   os.path.join(output_dir, "wandb.json"),
                   sort_keys=True,
                   indent=2)
        # store general configs
        wandb_run.config.update(prefix_dict(dict(gin_files=gin_files,
                                                 gin_bindings=gin_bindings,
                                                 output_dir=output_dir,
                                                 gpu=gpu), prefix='cli/'))
        wandb_run.config.update(note_params_dict)

    if remote_dir:
        import time
        logger.info("Test file upload to: {}".format(remote_dir))
        upload_dir(output_dir, remote_dir)
    return train(output_dir=output_dir, remote_dir=remote_dir, cometml_experiment=cometml_experiment, wandb_run=wandb_run)



import collections
import sys
import json
import numpy as np


class Logger(object):
    """tee functionality in python. If this object exists,
    then all of stdout gets logged to the file
    Adoped from:
    https://stackoverflow.com/questions/616645/how-do-i-duplicate-sys-stdout-to-a-log-file-in-python/3423392#3423392
    """

    def __init__(self, name, mode='a'):
        self.file = open(name, mode)
        self.stdout = sys.stdout
        sys.stdout = self

    def __del__(self):
        sys.stdout = self.stdout
        self.file.close()

    def write(self, data):
        self.file.write(data)
        self.stdout.write(data)
        # flush right away
        self.file.flush()
        self.stdout.flush()

    def flush(self):
        self.file.flush()
        self.stdout.flush()


class NumpyAwareJSONEncoder(json.JSONEncoder):

    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        elif isinstance(obj, np.generic):
            return obj.item()
        return json.JSONEncoder.default(self, obj)


def write_json(obj, fname, **kwargs):
    with open(fname, "w") as f:
        return json.dump(obj, f, cls=NumpyAwareJSONEncoder, **kwargs)


def prefix_dict(d, prefix=''):
    return {prefix + k: v for k, v in d.items()}



"""
Module implementing different samplers
"""
import pandas as pd
import numpy as np
from kipoi.external.torch.sampler import Sampler
from kipoi.data_utils import iterable_cycle
import warnings
import gin


def get_batch_sizes(p_vec, batch_size, verbose=True):
    """Compute the individual batch sizes for different probabilities

    Args:
      p_vec: list of probabilities for each class
      batch_size: batch size

    Returns:
      rounded list p_vec[i]*batch_size
    """
    p_vec = np.array(p_vec) / sum(p_vec)

    batch_sizes = np.round(p_vec * batch_size).astype(int)
    difference = batch_size - batch_sizes.sum()
    # Increase the largest one
    batch_sizes[batch_sizes.argmax()] += difference
    if verbose:
        print("Using batch sizes:")
        print(batch_sizes)
    assert batch_sizes.sum() == batch_size
    return batch_sizes


class StratifiedRandomBatchSampler(Sampler):

    def __init__(self, class_vector, p_vec, batch_size, verbose=False):
        """Stratified Sampling

        Args:
          class_vector (np.array): a vector of class labels
          p_vec (list[float]): list of probabilities for each class
          batch_size (int): batch_size
          verbose
        """
        self.n_splits = int(class_vector.shape[0] / batch_size)
        self.class_vector = class_vector
        self.p_vec = p_vec
        self.batch_size = batch_size

        self.batch_sizes = get_batch_sizes(self.p_vec, self.batch_size, verbose=verbose)

        # check that the individual batch size will always be > 0
        for i, batch_size in enumerate(self.batch_sizes):
            if batch_size == 0:
                warnings.warn("Batch size for class {} is 0.".format(i))

        self.classes = np.arange(len(p_vec))
        assert np.all(np.sort(pd.Series(self.class_vector).unique()) == self.classes)

        idx_all = np.arange(len(self.class_vector))
        self.class_idx_iterators = [iterable_cycle(np.random.permutation(idx_all[self.class_vector == cls]))
                                    for cls in self.classes]

    def __iter__(self):
        for i in range(len(self)):
            yield [next(self.class_idx_iterators[i])
                   for i, batch_size in enumerate(self.batch_sizes)
                   for j in range(batch_size)]

    def __len__(self):
        return len(self.class_vector) // self.batch_size



import sklearn.metrics as skm
import pandas as pd
import numpy as np
from collections import OrderedDict
import gin

# --------------------------------------------
# Combined metrics


@gin.configurable
class BootstrapMetric:
    def __init__(self, metric, n):
        """
        Args:
          metric: a function accepting (y_true and y_pred) and
             returning the evaluation result
          n: number of bootstrap samples to draw
        """
        self.metric = metric
        self.n = n

    def __call__(self, y_true, y_pred):
        outl = []
        for i in range(self.n):
            bsamples = (
                pd.Series(np.arange(len(y_true))).sample(frac=1, replace=True).values
            )
            outl.append(self.metric(y_true[bsamples], y_pred[bsamples]))
        return outl


@gin.configurable
class MetricsList:
    """Wraps a list of metrics into a single metric returning a list"""

    def __init__(self, metrics):
        self.metrics = metrics

    def __call__(self, y_true, y_pred):
        return [metric(y_true, y_pred) for metric in self.metrics]


@gin.configurable
class MetricsDict:
    """Wraps a dictionary of metrics into a single metric returning a dictionary"""

    def __init__(self, metrics):
        self.metrics = metrics

    def __call__(self, y_true, y_pred):
        return {k: metric(y_true, y_pred) for k, metric in self.metrics.items()}


@gin.configurable
class MetricsTupleList:
    """Wraps a dictionary of metrics into a single metric returning a dictionary"""

    def __init__(self, metrics):
        self.metrics = metrics

    def __call__(self, y_true, y_pred):
        return [(k, metric(y_true, y_pred)) for k, metric in self.metrics]


@gin.configurable
class MetricsOrderedDict:
    """Wraps a OrderedDict/tuple list of metrics into a single metric
    returning an OrderedDict
    """

    def __init__(self, metrics):
        self.metrics = metrics

    def __call__(self, y_true, y_pred):
        return OrderedDict([(k, metric(y_true, y_pred)) for k, metric in self.metrics])


@gin.configurable
class MetricsMultiTask:
    """Run the same metric across multiple tasks
    """

    def __init__(self, metrics, task_names=None):
        self.metrics = metrics
        self.task_names = task_names

    def __call__(self, y_true, y_pred):
        n_tasks = y_true.shape[1]
        if self.task_names is None:
            self.task_names = [i for i in range(n_tasks)]
        else:
            assert len(self.task_names) == n_tasks
        return OrderedDict([(task, self.metrics(y_true[:, i], y_pred[:, i]))
                            for i, task in enumerate(self.task_names)])


@gin.configurable
class MetricsAggregated:

    def __init__(self,
                 metrics,
                 agg_fn={"mean": np.mean, "std": np.std},
                 prefix=""):
        self.metrics
        self.agg_fn = agg_fn
        self.prefix = prefix

    def __call__(self, y_true, y_pred):
        out = self.metrics(y_true, y_pred)
        # TODO - generalize using numpy_collate?
        m = np.array(list(out.values()))
        return {self.prefix + k: fn(m) for k, fn in self.agg_fn}


@gin.configurable
class MetricsConcise:

    def __init__(self, metrics):
        import concise
        self.metrics_dict = OrderedDict([(m, concise.eval_metrics.get(m))
                                         for m in metrics])

    def __call__(self, y_true, y_pred):
        return OrderedDict([(m, fn(y_true, y_pred))
                            for m, fn in self.metrics_dict.items()])


# -----------------------------
# Binary classification
# Metric helpers
MASK_VALUE = -1
# Binary classification


def _mask_nan(y_true, y_pred):
    mask_array = ~np.isnan(y_true)
    if np.any(np.isnan(y_pred)):
        print("WARNING: y_pred contains {0}/{1} np.nan values. removing them...".
              format(np.sum(np.isnan(y_pred)), y_pred.size))
        mask_array = np.logical_and(mask_array, ~np.isnan(y_pred))
    return y_true[mask_array], y_pred[mask_array]


def _mask_value(y_true, y_pred, mask=MASK_VALUE):
    mask_array = y_true != mask
    return y_true[mask_array], y_pred[mask_array]


def _mask_value_nan(y_true, y_pred, mask=MASK_VALUE):
    y_true, y_pred = _mask_nan(y_true, y_pred)
    return _mask_value(y_true, y_pred, mask)


@gin.configurable
def n_positive(y_true, y_pred):
    return y_true.sum()


@gin.configurable
def n_negative(y_true, y_pred):
    return (1 - y_true).sum()


@gin.configurable
def frac_positive(y_true, y_pred):
    return y_true.mean()


@gin.configurable
def accuracy(y_true, y_pred, round=True):
    """Classification accuracy
    """
    y_true, y_pred = _mask_value_nan(y_true, y_pred)
    if round:
        y_true = np.round(y_true)
        y_pred = np.round(y_pred)
    return skm.accuracy_score(y_true, y_pred)


@gin.configurable
def auc(y_true, y_pred, round=True):
    """Area under the ROC curve
    """
    y_true, y_pred = _mask_value_nan(y_true, y_pred)

    if round:
        y_true = y_true.round()
    if len(y_true) == 0 or len(np.unique(y_true)) < 2:
        return np.nan
    return skm.roc_auc_score(y_true, y_pred)


@gin.configurable
def auprc(y_true, y_pred):
    """Area under the precision-recall curve
    """
    y_true, y_pred = _mask_value_nan(y_true, y_pred)
    return skm.average_precision_score(y_true, y_pred)


@gin.configurable
def mcc(y_true, y_pred, round=True):
    """Matthews correlation coefficient
    """
    y_true, y_pred = _mask_value_nan(y_true, y_pred)
    if round:
        y_true = np.round(y_true)
        y_pred = np.round(y_pred)
    return skm.matthews_corrcoef(y_true, y_pred)


@gin.configurable
def f1(y_true, y_pred, round=True):
    """F1 score: `2 * (p * r) / (p + r)`, where p=precision and r=recall.
    """
    y_true, y_pred = _mask_value_nan(y_true, y_pred)
    if round:
        y_true = np.round(y_true)
        y_pred = np.round(y_pred)
    return skm.f1_score(y_true, y_pred)


@gin.configurable
def cat_acc(y_true, y_pred):
    """Categorical accuracy
    """
    return np.mean(y_true.argmax(axis=1) == y_pred.argmax(axis=1))


classification_metrics = [
    ("auPR", auprc),
    ("auROC", auc),
    ("accuracy", accuracy),
    ("n_positive", n_positive),
    ("n_negative", n_negative),
    ("frac_positive", frac_positive),
]


@gin.configurable
class ClassificationMetrics:
    """All classification metrics
    """
    cls_metrics = classification_metrics

    def __init__(self):
        self.classification_metric = MetricsOrderedDict(self.cls_metrics)

    def __call__(self, y_true, y_pred):
        return self.classification_metric(y_true, y_pred)
# TODO - add gin macro for a standard set of classification and regession metrics


# --------------------------------------------
# Regression

@gin.configurable
def cor(y_true, y_pred):
    """Compute Pearson correlation coefficient.
    """
    y_true, y_pred = _mask_nan(y_true, y_pred)
    return np.corrcoef(y_true, y_pred)[0, 1]


@gin.configurable
def kendall(y_true, y_pred, nb_sample=100000):
    """Kendall's tau coefficient, Kendall rank correlation coefficient
    """
    from scipy.stats import kendalltau
    y_true, y_pred = _mask_nan(y_true, y_pred)
    if len(y_true) > nb_sample:
        idx = np.arange(len(y_true))
        np.random.shuffle(idx)
        idx = idx[:nb_sample]
        y_true = y_true[idx]
        y_pred = y_pred[idx]
    return kendalltau(y_true, y_pred)[0]


@gin.configurable
def mad(y_true, y_pred):
    """Median absolute deviation
    """
    y_true, y_pred = _mask_nan(y_true, y_pred)
    return np.mean(np.abs(y_true - y_pred))


@gin.configurable
def rmse(y_true, y_pred):
    """Root mean-squared error
    """
    return np.sqrt(mse(y_true, y_pred))


@gin.configurable
def rrmse(y_true, y_pred):
    """1 - rmse
    """
    return 1 - rmse(y_true, y_pred)


@gin.configurable
def mse(y_true, y_pred):
    """Mean squared error
    """
    y_true, y_pred = _mask_nan(y_true, y_pred)
    return ((y_true - y_pred) ** 2).mean(axis=None)


@gin.configurable
def ermse(y_true, y_pred):
    """Exponentiated root-mean-squared error
    """
    return 10**np.sqrt(mse(y_true, y_pred))


@gin.configurable
def var_explained(y_true, y_pred):
    """Fraction of variance explained.
    """
    y_true, y_pred = _mask_nan(y_true, y_pred)
    var_resid = np.var(y_true - y_pred)
    var_y_true = np.var(y_true)
    return 1 - var_resid / var_y_true


@gin.configurable
def pearsonr(y_true, y_pred):
    from scipy.stats import pearsonr
    y_true, y_pred = _mask_nan(y_true, y_pred)
    return pearsonr(y_true, y_pred)[0]


@gin.configurable
def spearmanr(y_true, y_pred):
    from scipy.stats import spearmanr
    y_true, y_pred = _mask_nan(y_true, y_pred)
    return spearmanr(y_true, y_pred)[0]


regression_metrics = [
    ("mse", mse),
    ("var_explained", var_explained),
    ("pearsonr", pearsonr),  # pearson and spearman correlation
    ("spearmanr", spearmanr),
    ("mad", mad),  # median absolute deviation
]


@gin.configurable
class RegressionMetrics:
    """All classification metrics
    """
    cls_metrics = regression_metrics

    def __init__(self):
        self.regression_metric = MetricsOrderedDict(self.cls_metrics)

    def __call__(self, y_true, y_pred):
        # squeeze the last dimension
        if y_true.ndim == 2 and y_true.shape[1] == 1:
            y_true = np.ravel(y_true)
        if y_pred.ndim == 2 and y_pred.shape[1] == 1:
            y_pred = np.ravel(y_pred)

        return self.regression_metric(y_true, y_pred)


# available eval metrics --------------------------------------------


BINARY_CLASS = ["auc", "auprc", "accuracy", "tpr", "tnr", "f1", "mcc"]
CATEGORY_CLASS = ["cat_acc"]
REGRESSION = ["mse", "mad", "cor", "ermse", "var_explained"]

AVAILABLE = BINARY_CLASS + CATEGORY_CLASS + REGRESSION



import os
import numpy as np
import pandas as pd
import logging
import sklearn

from gin_train.utils import write_json, prefix_dict
import gin

from tqdm import tqdm
from collections import OrderedDict
from abc import ABCMeta, abstractmethod

from kipoi.data_utils import numpy_collate_concat
from kipoi.external.flatten_json import flatten

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


@gin.configurable
class KerasTrainer:
    """Simple Keras model trainer
    """

    def __init__(self, model, train_dataset, valid_dataset, output_dir,
                 cometml_experiment=None, wandb_run=None):
        """
        Args:
          model: compiled keras.Model
          train: training Dataset (object inheriting from kipoi.data.Dataset)
          valid: validation Dataset (object inheriting from kipoi.data.Dataset)
          output_dir: output directory where to log the training
          cometml_experiment: if not None, append logs to commetml
        """
        self.model = model
        self.train_dataset = train_dataset
        self.valid_dataset = valid_dataset
        self.cometml_experiment = cometml_experiment
        self.wandb_run = wandb_run
        self.metrics = dict()

        if not isinstance(self.valid_dataset, list):
            # package the validation dataset into a list of validation datasets
            self.valid_dataset = [('valid', self.valid_dataset)]

        # setup the output directory
        self.output_dir = output_dir
        os.makedirs(self.output_dir, exist_ok=True)
        self.ckp_file = f"{self.output_dir}/model.h5"
        if os.path.exists(self.ckp_file):
            raise ValueError(f"model.h5 already exists in {self.output_dir}")
        self.history_path = f"{self.output_dir}/history.csv"
        self.evaluation_path = f"{self.output_dir}/evaluation.valid.json"

    def train(self,
              batch_size=256,
              epochs=100,
              early_stop_patience=4,
              num_workers=8,
              train_epoch_frac=1.0,
              valid_epoch_frac=1.0,
              train_samples_per_epoch=None,
              validation_samples=None,
              train_batch_sampler=None,
              tensorboard=True):
        """Train the model
        Args:
          batch_size:
          epochs:
          patience: early stopping patience
          num_workers: how many workers to use in parallel
          train_epoch_frac: if smaller than 1, then make the epoch shorter
          valid_epoch_frac: same as train_epoch_frac for the validation dataset
          train_batch_sampler: batch Sampler for training. Useful for say Stratified sampling
          tensorboard: if True, tensorboard output will be added
        """
        from keras.callbacks import EarlyStopping, History, CSVLogger, ModelCheckpoint, TensorBoard
        from keras.models import load_model

        if train_batch_sampler is not None:
            train_it = self.train_dataset.batch_train_iter(shuffle=False,
                                                           batch_size=1,
                                                           drop_last=None,
                                                           batch_sampler=train_batch_sampler,
                                                           num_workers=num_workers)
        else:
            train_it = self.train_dataset.batch_train_iter(batch_size=batch_size,
                                                           shuffle=True,
                                                           num_workers=num_workers)
        next(train_it)
        valid_dataset = self.valid_dataset[0][1]  # take the first one
        valid_it = valid_dataset.batch_train_iter(batch_size=batch_size,
                                                  shuffle=True,
                                                  num_workers=num_workers)
        next(valid_it)

        if tensorboard:
            tb = [TensorBoard(log_dir=self.output_dir)]
        else:
            tb = []

        if self.wandb_run is not None:
            from wandb.keras import WandbCallback
            wcp = [WandbCallback(save_model=False)]  # we save the model using ModelCheckpoint
        else:
            wcp = []

        # train the model
        if len(valid_dataset) == 0:
            raise ValueError("len(self.valid_dataset) == 0")

        if train_samples_per_epoch is None:
            train_steps_per_epoch = max(int(len(self.train_dataset) / batch_size * train_epoch_frac), 1)
        else:
            train_steps_per_epoch = max(int(train_samples_per_epoch / batch_size), 1)
            
        
        if validation_samples is None:
            # parametrize with valid_epoch_frac
            validation_steps = max(int(len(valid_dataset) / batch_size * valid_epoch_frac), 1)
        else:
            validation_steps = max(int(validation_samples / batch_size), 1)

        self.model.fit_generator(train_it,
                                 epochs=epochs,
                                 steps_per_epoch=train_steps_per_epoch,
                                 validation_data=valid_it,
                                 validation_steps=validation_steps,
                                 callbacks=[EarlyStopping(patience=early_stop_patience,
                                                          restore_best_weights=True),
                                            CSVLogger(self.history_path)] + tb + wcp
                                            # ModelCheckpoint(self.ckp_file, save_best_only=True)] 
                                 )
        self.model.save(self.ckp_file)
        # self.model = load_model(self.ckp_file)  # not necessary, EarlyStopping is already restoring the best weights

        # log metrics from the best epoch
        dfh = pd.read_csv(self.history_path)
        m = dict(dfh.iloc[dfh.val_loss.idxmin()])
        if self.cometml_experiment is not None:
            self.cometml_experiment.log_multiple_metrics(m, prefix="best-epoch/")
        if self.wandb_run is not None:
            self.wandb_run.summary.update(flatten(prefix_dict(m, prefix="best-epoch/"), separator='/'))
        self.metrics = flatten(prefix_dict(m, prefix="best-epoch/"))

    #     def load_best(self):
    #         """Load the best model from the Checkpoint file
    #         """
    #         self.model = load_model(self.ckp_file)

    def evaluate(self, metric, batch_size=256, num_workers=8, eval_train=False, eval_skip=False, save=True, **kwargs):
        """Evaluate the model on the validation set
        Args:
          metrics: a list or a dictionary of metrics
          batch_size:
          num_workers:
          eval_train: if True, also compute the evaluation metrics on the training set
          save: save the json file to the output directory
        """
        if len(kwargs) > 0:
            logger.warn(f"Extra kwargs were provided to trainer.evaluate: {kwargs}")
        # contruct a list of dataset to evaluate
        if eval_train:
            eval_datasets = [('train', self.train_dataset)] + self.valid_dataset
        else:
            eval_datasets = self.valid_dataset

        try:
            if len(eval_skip) > 0:
                eval_datasets = [(k,v) for k,v in eval_datasets if k not in eval_skip]
        except:
            logger.warn(f"eval datasets don't contain tuples. Unable to skip them using {eval_skip}")

        metric_res = OrderedDict()
        for d in eval_datasets:
            if len(d) == 2:
                dataset_name, dataset = d
                eval_metric = metric  # use the default eval metric
            elif len(d) == 3:
                # specialized evaluation metric was passed
                dataset_name, dataset, eval_metric = d
            else:
                # TODO - this should be made more explicit with classes
                raise ValueError("Valid dataset needs to be a list of tuples of 2 or 3 elements"
                                 "(name, dataset) or (name, dataset, metric)")
            logger.info(f"Evaluating dataset: {dataset_name}")
            lpreds = []
            llabels = []
            from copy import deepcopy
            for inputs, targets in tqdm(dataset.batch_train_iter(cycle=False,
                                                                 num_workers=num_workers,
                                                                 batch_size=batch_size),
                                        total=len(dataset) // batch_size
                                        ):
                lpreds.append(self.model.predict_on_batch(inputs))
                llabels.append(deepcopy(targets))
                del inputs
                del targets
            preds = numpy_collate_concat(lpreds)
            labels = numpy_collate_concat(llabels)
            del lpreds
            del llabels
            metric_res[dataset_name] = eval_metric(labels, preds)

        if save:
            write_json(metric_res, self.evaluation_path, indent=2)
            logger.info("Saved metrics to {}".format(self.evaluation_path))

        if self.cometml_experiment is not None:
            self.cometml_experiment.log_multiple_metrics(flatten(metric_res, separator='/'), prefix="eval/")

        if self.wandb_run is not None:
            self.wandb_run.summary.update(flatten(prefix_dict(metric_res, prefix="eval/"), separator='/'))
        metric_res = {**self.metrics, **metric_res}
        return metric_res
    

class Trainer(object):
    """Generic trainer object. This class has two major components:
            (i) train model
            (ii) evalute model (this implies also predicting using the model)
    
        
    Attributes:
        model: compiled sklearn.pipeline.Pipeline
        train (:obj:RepDataset): training dataset 
                                The object should have the following properties: 
                                obj.inputs, obj.targets, obj.metadata, obj.dataset_name
        valid (:obj:RepDataset): validation dataset 
        output_dir: output directory where to log the training
        cometml_experiment: if not None, append logs to commetml
        wandb_run: send evaluation scores to the dashbord
    """
    def __init__(self,
                 model,
                 train_dataset,
                 valid_dataset,
                 output_dir,
                 cometml_experiment=None,
                 wandb_run=None):
        
        self.model = model
        self.train_dataset = train_dataset
        self.valid_dataset = valid_dataset
        self.cometml_experiment = cometml_experiment
        self.wandb_run = wandb_run

        # setup the output directory
        self.set_output_dir(output_dir)


    def set_output_dir(self, output_dir):
        """Set output folder structure for the model.

        Args:
            output_dir (str): output directory name
        """

        self.output_dir = output_dir
        os.makedirs(self.output_dir, exist_ok=True)
        self.ckp_file = f"{self.output_dir}/model.h5"
        if os.path.exists(self.ckp_file):
            raise ValueError(f"model.h5 already exists in {self.output_dir}")
        self.history_path = f"{self.output_dir}/history.csv"
        self.evaluation_path = f"{self.output_dir}/evaluation.valid.json"
    
    
      
    ##########################################################################
    ###########################    Train   ###################################
    ##########################################################################
    
    def train(self,
              batch_size=256,
              epochs=100,
              early_stop_patience=4,
              num_workers=8,
              train_epoch_frac=1.0,
              valid_epoch_frac=1.0,
              train_samples_per_epoch=None,
              validation_samples=None,
              train_batch_sampler=None,
              **kwargs):

        # **kwargs won't be used, they are just included for compatibility with gin_train.
        """Train the model
        
        Args:
            num_workers: how many workers to use in parallel
        """
        
        # define dataset
        X_train, y_train = (self.train_dataset.inputs, self.train_dataset.targets)
        if self.valid_dataset is None:
            raise ValueError("len(self.valid_dataset) == 0")

        # check model type
        self.check_model()

        # fit model
        self.fit(X_train, y_train, epochs=epochs, batch_size=batch_size, num_workers=num_workers)
        
        # save model
        self.save()
        
    
    @abstractmethod
    def check_model(self):
        """Check if the model has the specified data type."""
        pass
    
    
    @abstractmethod
    def fit(self, inputs, targets):
        """Generic method for fitting using a given model.
           This method has to be implemented in subclass
        """
        pass
    
    
    @abstractmethod
    def save(self):
        """Generic method for saving the model.
           This method has to be implemented in subclass.
        """
        pass
    
    
    ##########################################################################
    ########################### Evaluation ###################################
    ##########################################################################
    

    def evaluate(self, 
                 metric, 
                 batch_size = None,
                 num_workers = 8,
                 eval_train=False,
                 eval_skip=False,
                 save=True):
        """Evaluate the model on the validation set

        Args:
            metric: a list or a dictionary of metrics
            batch_size: None - means full dataset
            num_workers: number of threads
            eval_train: if True, also compute the evaluation metrics on the training set
            save: save the json file to the output directory
        """
        
        # contruct a list of dataset to evaluate
        if eval_train:
            eval_datasets = [self.train_dataset, self.valid_dataset]
        else:
            eval_datasets = [self.valid_dataset]
        
        metric_res = OrderedDict()
        eval_metric = metric
        
        for i, data in enumerate(eval_datasets):

            lpreds = []
            llabels = []
            
            inputs = data.inputs
            targets = data.targets
            
            lpreds.append(self.predict(inputs))
            llabels.append(deepcopy(targets))

            preds = numpy_collate_concat(lpreds)
            labels = numpy_collate_concat(llabels)
            del lpreds
            del llabels
            
            metric_res[data.dataset_name] = eval_metric(labels, preds, tissue_specific_metadata = data.metadata)
                
        if save:
            write_json(metric_res, self.evaluation_path, indent=2)
            logger.info("Saved metrics to {}".format(self.evaluation_path))

        if self.cometml_experiment is not None:
            self.cometml_experiment.log_multiple_metrics(flatten(metric_res, separator='/'), prefix="eval/")

        if self.wandb_run is not None:
            self.wandb_run.summary.update(flatten(prefix_dict(metric_res, prefix="eval/"), separator='/'))

        return metric_res
    
    
    @abstractmethod
    def predict(self, inputs):
        """Generic method for predicting using a given model.
           This method has to be implemented in subclass.
        """
        pass


@gin.configurable
class SklearnPipelineTrainer(Trainer):

    """Simple Scikit model trainer
    """

    def __init__(self,
                 model,
                 train_dataset,
                 valid_dataset,
                 output_dir,
                 cometml_experiment=None,
                 wandb_run=None):

        Trainer.__init__(self,model,
                 train_dataset,
                 valid_dataset,
                 output_dir,
                 cometml_experiment,
                 wandb_run)

    
    def check_model(self):
        if not isinstance(self.model, sklearn.pipeline.Pipeline):
            raise ValueError("model is not a sklearn.pipeline.Pipeline")
    
    
    def fit(self, inputs, targets, epochs=10, batch_size=256, num_workers=8):        
        self.model.fit(inputs, targets)
    
    
    def save(self):        
        import pickle
        with open(self.ckp_file, 'wb') as file:
            pickle.dump(self.model, file)


    def predict(self, inputs):
        return self.model.predict(inputs)



"""Test fs
"""
from fs.osfs import OSFS
from fs.copy import copy_file_if_newer
from fs import open_fs
import os
from tqdm import tqdm


def _open_fs(directory):
    if directory.startswith("s3://"):
        """Manually fetch the permissions from the environment

        Requires the following env variables:
        - S3_ACCESS_KEY
        - S3_SECRET_KEY
        - S3_URL
        """
        from fs_s3fs import S3FS
        if not directory.endswith("/"):
            directory += "/"

        bucket, fpath = directory[len("s3://"):].split("/", 1)
        return S3FS(bucket, dir_path=fpath,
                    aws_access_key_id=os.environ.get("S3_ACCESS_KEY", None),
                    aws_secret_access_key=os.environ.get('S3_SECRET_KEY', None),
                    strict=False,
                    endpoint_url=os.environ.get('S3_URL', None))
    else:
        return open_fs(directory)


def upload_dir(local_dir, remote_dir):
    """Upload directory to the remote directory

    Args:
      local_dir: local directory
      remote_dir: remote directory

    Returns:
      List of urls
    """
    remote_fs = _open_fs(remote_dir)
    local_fs = OSFS(local_dir)
    # urls = []
    for f in tqdm(list(local_fs.walk.files())):
        copy_file_if_newer(local_fs, f, remote_fs, f)
        # urls.append((f, remote_fs.geturl(f)))
    # return urls


# upload_dir("tests", 's3://bucket1/asdas/folder2')





def create_tf_session(visiblegpus, per_process_gpu_memory_fraction=0.45):
    import os
    import tensorflow as tf
    import keras.backend as K
    os.environ['CUDA_VISIBLE_DEVICES'] = str(visiblegpus)
    session_config = tf.ConfigProto()
    # session_config.gpu_options.deferred_deletion_bytes = DEFER_DELETE_SIZE
    session_config.gpu_options.per_process_gpu_memory_fraction = per_process_gpu_memory_fraction
    session = tf.Session(config=session_config)
    K.set_session(session)
    return session



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

requirements = [
    "kipoi>=0.6.0",
    "gin-config",
    "comet_ml",
    "numpy",
    "pandas",
    "tqdm",
    "colorlog",
    "argh",
    "scikit-learn",
    "fs",
    "fs-s3fs",
    # sometimes required
]

test_requirements = [
    "bumpversion",
    "wheel",
    "epc",
    "jedi",
    "pytest>=3.3.1",
    "pytest-xdist",  # running tests in parallel
    "pytest-pep8",
    "pytest-cov",
    "coveralls",
]

setup(
    name='gin-train',
    version='0.1.17',
    description="gin-train: model training boilerplate",
    author="Ziga Avsec",
    author_email='avsec@in.tum.de',
    url='https://github.com/avsecz/gin-train',
    long_description="gin-train: model training boilerplate",
    packages=find_packages(),
    install_requires=requirements,
    extras_require={
        "develop": test_requirements,
    },
    entry_points={'console_scripts': ['gin_train = gin_train.__main__:main', 'gt = gin_train.__main__:main']},
    license="MIT license",
    zip_safe=False,
    keywords=["deep learning"],
    test_suite='tests',
    package_data={'gin_train': ['logging.conf']},
    tests_require=test_requirements
)




from __future__ import absolute_import

__author__ = 'Ziga Avsec'
__email__ = 'avsec@in.tum.de'
__version__ = '0.1.17'

import comet_ml  # need to import this
from . import metrics
from . import trainers
from . import utils
from . import config
from . import cli
